% Grid-restrained simplex algorithm implementation
% (C)2006-2011 Arpad Buermen
% Distributed under the terms of LGPL v2.1. 
%
% function [x,f,niter]=gridnm(cf, s0, dorigin0, d0, nmax, fabstol, xabstol, reltol, gridsize)
%
% Return values
%   x        ... solution
%   f        ... cf at solution
%   niter    ... number of cf evaluations
%
% Arguments
%   cf       ... function name
%   s0       ... initial simplex (columns represent simplex points)
%   dorigin0 ... initial grid origin
%   d0       ... initial grid scaling vector
%   nmax     ... maximal number of cf evaluations before the algorithm stops
%   fabstol  ... absolute f tolerance
%   xabstol  ... absolute x tolerance
%   reltol   ... relative tolerance

function [x,f,niter]=gridnm(cf, s0, dorigin0, d0, nmax, fabstol, xabstol, reltol)
% Coefficients
gammar=1.0;
gammae=1.2;
gammaco=0.5;
gammaci=-0.5;
gammap=1/4;
% Grid step 
%   Empty value means that there will be no rounding (floor) to the power of gridstep
gridstep=[];
% Bad shape factor
shapefact=1e-6;
% Grid size factor 
normfact=2.0;
%normfactmax=2^53;
normfactmax=2^52;   % difference
% Grid tolerance
gridreltol=2^-52;
gridabstol=1e-100;

% Set initial simplex
s=s0;
% Set initial grid
gridorigin=dorigin0;
gridsize=d0;
gridsize=gridsnap(gridsize, gridstep, dorigin0, gridreltol, gridabstol);
% Evaluate initial simplex
f=feval(cf, s);

% Initialize CF evaluation counter
niter=length(f);

% Main loop
while niter<nmax
    % Stat simplex
    [n,m]=size(s);
    % Sort simplex
    S=[f;s];
    Ss=sortrows(S')';
    f=Ss(1,1:end);
    s=Ss(2:end,1:end);

    % Get simplex sides
    v=Ss(2:end,2:end)-(Ss(2:end,1))*ones(1,n);
    % Get valley direction
    vnorm=sqrt(sum(v.*v,1));
    [maxl,maxi]=max(vnorm);
    
    % Get centroid
    xc=mean(s(1:end,1:end-1),2);
    % Get worst point
    xw=s(1:end,end);
    fw=f(end);
    % Get second worst point
    xsw=s(1:end,end-1);
    fsw=f(end-1);
    % Get best point
    xb=s(1:end,1);
    fb=f(1);
    % Set failure flag to 0
    failed=0;
    % Try reflection
    xr=xc-gammar*(xw-xc);
    xr=ptrx(xr,gridorigin,gridsize);
    fr=feval(cf,xr);
    niter=niter+1;
    if fr<fsw        
       if fb<=fr
           f(end)=fr;
           s(1:end,end)=xr;
       else
           % Try expansion
           xe=xc-gammae*(xw-xc);
           xe=ptrx(xe,gridorigin,gridsize);
           fe=feval(cf,xe);
           niter=niter+1;
           if fe<fr
               f(end)=fe;
               s(1:end,end)=xe;
           else
               f(end)=fr;
               s(1:end,end)=xr;
           end
       end
    else
        if fr>=fw
            % Try inner contraction
            xci=xc-gammaci*(xw-xc);
            xci=ptrx(xci,gridorigin,gridsize);
            fci=feval(cf,xci);
            niter=niter+1;
            if fci<fsw           % xxx
               f(end)=fci;
               s(1:end,end)=xci;
            else
               failed=1;
            end
        else
            % Try outer contraction
            xco=xc-gammaco*(xw-xc);
            xco=ptrx(xco,gridorigin,gridsize);
            fco=feval(cf,xco);
            niter=niter+1;
            if fco<fsw           % xxx
               f(end)=fco;
               s(1:end,end)=xco;
            else
               failed=1;
            end
        end
    end

    % Sort simplex
    [n,m]=size(s);
    S=[f;s];
    Ss=sortrows(S')';
    % Get simplex sides
    v=Ss(2:end,2:end)-(Ss(2:end,1))*ones(1,n);
    % Get side lengths
    vnorm=sqrt(sum(v.*v,1));
    % Get side divisions
    Nv=round(abs(v./(gridsize*ones(1,n))));
    Nvmax=max(Nv,[],2);
    % Check shape by means of QR decomp.
    % QR is (according to the algorithm) needed only if NM iteration fails
    % Here we do it all the time to dump the shape.
    Sv=sortrows([vnorm; v]')';
    v=Sv(2:end,end:-1:1);
    [Q,R]=qr(v);
    RD=diag(R);
    [rdmin, imin]=min(abs(RD));
    detv=rdmin/(shapefact*sqrt(n)*norm(gridsize)/2);
    % If NM iteration failed and the shape is bad, do a reshape in the probe loop
    
    if (detv<1) & (failed==1)
        badshape=1;
    else
        badshape=0;
    end
    
    % Dump NM step info
    if failed==1
        txt='FAIL';
    else
        txt='OK  ';
    end
    disp(sprintf('%s_%d: NM %s: n=%6d grid=%8.2g..%8.2g Nv=%8.2g..%8.2g det=%8.2g f=%.20g', cf, n, txt, niter, min(gridsize), max(gridsize), min(Nvmax), max(Nvmax), detv, min(f)));
 
    % Sort simplex
    [n,m]=size(s);
    S=[f;s];
    Ss=sortrows(S')';
    % Get simplex sides
    v=Ss(2:end,2:end)-(Ss(2:end,1))*ones(1,n);
    % Get side lengths
    vnorm=sqrt(sum(v.*v,1));
    [maxl,maxi]=max(vnorm);
    
    % Probe loop
    if failed==1
        % NM iteration failed
        [n,m]=size(s);

        % Sort simplex
        S=[f;s];
        Ss=sortrows(S')';
        f=Ss(1,1:end);
        s=Ss(2:end,1:end);
	
        % Get best point
        xb=s(:,1);
        fb=f(1);
        
        % Not reshaped in this probe cycle
        reshaped=0;
        
        % Pseudo-expand, assume it failed
        pefailed=1;

        % Sort simplex
        S=[f;s];
        Ss=sortrows(S')';
        f=Ss(1,1:end);
        s=Ss(2:end,1:end);

        % Get best point
        xb=s(:,1);
        fb=f(1);

        % If needed, do a reshape
        if badshape==1
            % Do a reshape
            disp('Reshape before PE');
            vresh=Q;
            sclresh=RD;
            % Clip base
            baseresh=clipbasis(vresh, sclresh, normfact*sqrt(n)*norm(gridsize)/2, normfactmax*sqrt(n)*norm(gridsize)/2);
            frmpts=m;
            % Reshaped base vectors and the pseudo-expand vector
            dresh=[baseresh,-sum(baseresh,2)/n*(gammae-gammar)/gammar];
            % Get and evaluate the frame points
            xnew=xb*ones(1,frmpts)+dresh;
            xnew=ptrx(xnew,gridorigin,gridsize);
            fnew=feval(cf, xnew);
            % xnew and fnew contain the pseudo-expand point and all points except the origin (original best)
            niter=niter+frmpts;
            % Signal that the simplex was reshaped
            reshaped=1;
        else
            % No reshape needed, just evaluate pseudo-expand
            xcw=sum(s(:,2:end),2)/n;
            xpe=xb+(gammae-gammar)/gammar*(xb-xcw);
            xpe=ptrx(xpe,gridorigin,gridsize);
            fpe=feval(cf, xpe);
            niter=niter+1;
            % xnew and fnew contain the pseudo-expand point and all points except the origin (original best)
            xnew=[s(:,2:end),xpe];
            fnew=[f(:,2:end),fpe];
        end

        % Reshape leaves the best point (xb, fb) unchanged
        if min(fnew)<fb
            % Pseudo-expand and reshape OK, best point changed
            pefailed=0;
            disp(sprintf('%s_%d: PE OK  : n=%6d f=%.20g', cf, n, niter, min(fnew)));
        else
            % Pseudo-expand and reshape failed to produce a better point
            disp(sprintf('%s_%d: PE FAIL: n=%6d f=%.20g', cf, n, niter, min(fnew)));
        end
        
        if pefailed==1
            % Pseudo-expand failed
	
            % Probe loop
            ncredit=0;
            nprobe=0;
            if reshaped
                % If the base is reshaped, one base is already evaluated
                nprobe=1;
            else
                % If it is not reshaped, no bases are evaluated yet
                nprobe=0;
            end
            while min(fnew)>=fb
                % If pseudo-expand didn't reshape the base, consider reshaping
                if ~reshaped
                    vresh=Q;
                    sclresh=RD;
                    disp('Reshape');
                    baseresh=clipbasis(vresh, sclresh, normfact*sqrt(n)*norm(gridsize)/2, normfactmax*sqrt(n)*norm(gridsize)/2);
                    % Form the frame template vector set
                    frmpts=m-1;
                    dresh=[baseresh];
                    % Remember that the base is reshaped
                    reshaped=1;
                else
                    % Remove the pseudo-expand point as we don't want it in the probe loop
                    frmpts=m-1;
                    dresh=dresh(:,1:frmpts);
                    % Shrink only after 2 bases have been evaluated and the number of base evaluations is even
                    if (nprobe>=2) & (mod(nprobe,2)==0)
                        % Shrink only if both dresh and -dresh were examined
                        % Shrink and mirror frame
                        dresh=-dresh*gammap;
                        % Shrink scale factors
                        % sclresh=sclresh*gammap;
                        % Set the number of allowed grid refinements
                        ncredit=1;
                    else
                        % Only mirror frame
                        dresh=-dresh;
                    end % if shrink
                end

                % Check grid
                if ncredit>0
                    % We are allowed to change the grid (2 probe steps failed, orthogonal maximal positive base examined)
                    [lmin, imin]=min(sqrt(sum(dresh.*dresh, 1)));
                    [lmax, imax]=max(sqrt(sum(dresh.*dresh, 1)));
                    vmin=dresh(:,imin);
                    vmax=dresh(:,imax);
                    % Do we need a finer grid
                    if (lmin<normfact*sqrt(n)*norm(gridsize)/2)
                        % Yes we do
                        
                        % Shortest side: division in any coordinate direction should be less than lmin/n
                        gshort=max(abs(vmin)/(250*normfact*n), lmin/(250*normfact*n^1.5));
                        
                        % Longest side should have at least 500 div per direction
                        glong=gridsize;
                         
                        % Choose minimal grid step for every direction
                        gridsizenew=min([gshort, glong, gridsize],[],2);
                        gridorigin=xb;
                        
                        % Limit the grid to machine precision and embed it in previous grid
                        [gridsize, clipped]=gridsnap(gridsizenew, gridstep, xb, gridreltol, gridabstol);
                        if sum(clipped)>0
                            txt='CLIP';
                        else
                            txt='    ';
                        end
	
                        % Dump info
                        disp(sprintf('%s_%d: SW %s: n=%6d grid=%8.2g..%8.2g f=%.20g', cf, n, txt, niter, min(gridsize), max(gridsize), min(f)));
                    end
                end
                
                % Evaluate probe points
                xnew=xb*ones(1,frmpts)+dresh;
                xnew=ptrx(xnew,gridorigin,gridsize);
                fnew=feval(cf, xnew);
                niter=niter+frmpts;

                % Get coordinate divisions per simplex side
                Nv=round(abs(dresh./(gridsize*ones(1,frmpts))));
                Nvmax=max(Nv,[],2);
                % Check for improvement
                if min(fnew)<fb
                    % Improvement of the best point, stop loop
                    txt='OK  ';
                else
                    txt='FAIL';
                end
                % Dump info
                disp(sprintf('%s_%d: PR %s: n=%6d grid=%8.2g..%8.2g Nv=%8.2g..%8.2g f=%.20g', cf, n, txt ,niter, min(gridsize), max(gridsize), min(Nvmax), max(Nvmax), min([fb, fnew])));

                % Increase the loop counter
                nprobe=nprobe+1;
                
                % Check tolerances
                if checkftol([fb, fnew], fabstol, reltol) & checkxtol([xb, xnew], [fb, fnew], xabstol, reltol)
                    % Within ftol and xtol, stop loop
                    break;
                end
            end % while
        end % if pefailed
        
        % Form a new simplex
        % xnew is either the reshaped simplex + pseudo-expand point or
        %         the original simplex + pseudo-expand point
        % m=n+1
        % xnew and fnew contain the pseudo-expand point and 
        %               all points except the origin (original best)
        % m-th point is pseudo-expand
        [dummy,imin]=min(fnew);
        if imin<m
            % Pseudo-expand is not the best
            s=[xb,xnew(:,1:m-1)];
            f=[fb,fnew(1:m-1)];
        else
            % Pseudo-expand is the best
            s=xnew;
            f=fnew;
        end
    end % if failed

    % Check tolerances
    xtolstop=checkxtol(s, f, xabstol, reltol);
    ftolstop=checkftol(f, fabstol, reltol);
    if xtolstop & ftolstop
        disp('Stopped');
        break;
    end
end
[f,imin]=min(f);
x=s(1:end,imin);
x=ptrx(x,gridorigin,gridsize);

function x1=ptrx(x,x0,d)
[nrow,ncol]=size(x);
x0l=x0*ones(1,ncol);
d1=d*ones(1,ncol);
n1=(x-x0l)./d1;
n=round(n1);
x1=x0l+n.*d1;

function [g, clipped]=gridsnap(gridsize, base, origin, reltol, abstol)
gmin=max(abs(origin)*reltol,(origin*0+1)*abstol);
clipped=gridsize<gmin;
g=max(gmin, gridsize);
if ~isempty(base)
    g=base.^floor(log(g)/log(base));
end

function [Vout]=clipbasis(Q, D, RminGrid, RmaxGrid)
    SD=sign(D);
    SD=SD+1*(SD==0);
    D=max(RminGrid, min(abs(D), RmaxGrid));
    Vout=Q*diag(SD.*D);

function within=checkftol(f, fabstol, reltol)
    % Check f tolerance
    fmin=min(f);
    fmax=max(f);
    fdif=fmax-fmin;
    tol=max(abs(fmin)*reltol, fabstol);
    within=0;
    if sum(fdif>=tol)<=0
        within=1;
    end
    
function within=checkxtol(s, f, xabstol, reltol)
    % Check x tolerance
    [n,m]=size(s);
    S=[f;s];
    Ss=sortrows(S')';
    f=Ss(1,1:end);
    s=Ss(2:end,1:end);
    xb=s(:,1);
    fb=f(1);
    d=s-xb*ones(1,m);
    tol=max(abs(xb)*reltol, xabstol);
    within=0;
    if sum(max(abs(d),[],2)>=tol)<=0
        within=1;
    end
